@@ -176,16 +176,17 @@ impl Frames {
     fn all_uniq(&self) -> Vec<Arc<FrameData>> {
         let mut all = self.slowest.clone();
         all.extend(self.recent.iter().cloned());
-        all.sort_by_key(|frame| frame.frame_index);
-        all.dedup_by_key(|frame| frame.frame_index);
+        all.sort_by_key(|frame| frame.frame_index());
+        all.dedup_by_key(|frame| frame.frame_index());
         all
     }
 }
 
 #[derive(Clone)]
 pub struct Paused {
-    /// What we are viewing
-    selected: Arc<FrameData>,
+    /// The frame we are viewing.
+    selected_frame: Arc<FrameData>,
+    /// All the frames we had when paused.
     frames: Frames,
 }
 
@@ -209,6 +210,10 @@ pub struct ProfilerUi {
 
     /// How we normalize the frame view:
     slowest_frame: f32,
+
+    /// When did we last run a pass to pack all the frames?
+    #[serde(skip)]
+    last_pack_pass: Option<std::time::Instant>,
 }
 
 impl Default for ProfilerUi {
@@ -223,6 +228,7 @@ impl Default for ProfilerUi {
             is_zooming: false,
             paused: None,
             slowest_frame: 0.17,
+            last_pack_pass: None,
         }
     }
 }
@@ -362,12 +368,12 @@ impl ProfilerUi {
     }
 
     /// Pause on the specific frame
-    fn pause_and_select(&mut self, selected: Arc<FrameData>) {
+    fn pause_and_select(&mut self, selected_frame: Arc<FrameData>) {
         if let Some(paused) = &mut self.paused {
-            paused.selected = selected;
+            paused.selected_frame = selected_frame;
         } else {
             self.paused = Some(Paused {
-                selected,
+                selected_frame,
                 frames: self.frames(),
             });
         }
@@ -376,12 +382,38 @@ impl ProfilerUi {
     fn selected_frame(&self) -> Option<Arc<FrameData>> {
         self.paused
             .as_ref()
-            .map(|paused| paused.selected.clone())
+            .map(|paused| paused.selected_frame.clone())
             .or_else(|| self.frame_view.lock().latest_frame())
     }
 
     fn selected_frame_index(&self) -> Option<FrameIndex> {
-        self.selected_frame().map(|frame| frame.frame_index)
+        self.selected_frame().map(|frame| frame.frame_index())
+    }
+
+    fn all_known_frames(&self) -> Vec<Arc<FrameData>> {
+        let mut all = self.frame_view.lock().all_uniq();
+        if let Some(paused) = &self.paused {
+            all.append(&mut paused.frames.all_uniq());
+        }
+        all.sort_by_key(|frame| frame.frame_index());
+        all.dedup_by_key(|frame| frame.frame_index());
+        all
+    }
+
+    fn run_pack_pass_if_needed(&mut self) {
+        let last_pack_pass = self
+            .last_pack_pass
+            .get_or_insert_with(std::time::Instant::now);
+        let time_since_last_pack = last_pack_pass.elapsed();
+        if time_since_last_pack > std::time::Duration::from_secs(1) {
+            puffin::profile_scope!("pack_pass");
+            for frame in self.all_known_frames() {
+                if Some(frame.frame_index()) != self.selected_frame_index() {
+                    frame.pack();
+                }
+            }
+            self.last_pack_pass = Some(std::time::Instant::now());
+        }
     }
 
     /// Show the profiler.
@@ -392,6 +424,8 @@ impl ProfilerUi {
 
         puffin::profile_function!();
 
+        self.run_pack_pass_if_needed();
+
         let mut scopes_on = puffin::are_scopes_on();
         ui.checkbox("Profiling enabled", &mut scopes_on);
         puffin::set_scopes_on(scopes_on);
@@ -411,17 +445,25 @@ impl ProfilerUi {
             ui.unindent();
         }
 
-        let frame = match hovered_frame.or_else(|| self.selected_frame()) {
-            Some(frame) => frame,
-            None => {
-                ui.text("No profiling data");
+        let frame = hovered_frame.or_else(|| self.selected_frame());
+        let frame = if let Some(frame) = frame {
+            frame
+        } else {
+            ui.text("No profiling data");
+            return;
+        };
+
+        let frame = match frame.unpacked() {
+            Ok(frame) => frame,
+            Err(err) => {
+                ui.text_colored(ERROR_COLOR, format!("Bad frame: {}", err));
                 return;
             }
         };
 
         // TODO: show age of data
 
-        let (min_ns, max_ns) = frame.range_ns;
+        let (min_ns, max_ns) = frame.range_ns();
 
         ui.button("Help!");
         if ui.is_item_hovered() {
@@ -464,12 +506,11 @@ impl ProfilerUi {
         );
 
         ui.text(format!(
-            "Showing frame #{}, {:.1} ms, {} threads, {} scopes, {:.1} kB",
-            frame.frame_index,
+            "Showing frame #{}, {:.1} ms, {} threads, {} scopes.",
+            frame.frame_index(),
             (max_ns - min_ns) as f64 * 1e-6,
             frame.thread_streams.len(),
-            frame.num_scopes,
-            frame.num_bytes as f64 * 1e-3
+            frame.meta.num_scopes,
         ));
 
         // The number of threads can change between frames, so always show this even if there currently is only one thread:
@@ -514,7 +555,7 @@ impl ProfilerUi {
     fn ui_canvas(
         &mut self,
         info: &Info<'_>,
-        frame: &Arc<FrameData>,
+        frame: &Arc<UnpackedFrameData>,
         (min_ns, max_ns): (NanoSecond, NanoSecond),
     ) -> f32 {
         puffin::profile_function!();
@@ -597,25 +638,7 @@ impl ProfilerUi {
 
         let mut hovered_frame = None;
 
-        {
-            let uniq = frames.all_uniq();
-            let bytes: usize = uniq.iter().map(|frame| frame.num_bytes).sum();
-            ui.text(format!(
-                "{} frames recorded ({:.1} MB)",
-                uniq.len(),
-                bytes as f64 * 1e-6
-            ));
-
-            let mut memory_length = self.frame_view.lock().max_recent() as u64;
-            imgui::Slider::new("Max frames to store", 10, 100_000)
-                .display_format(format!(
-                    "%d (~ {:.1} min at 60 Hz, ~ {:.0} MB)",
-                    memory_length as f32 / 60.0 / 60.0,
-                    memory_length as f32 * bytes as f32 / uniq.len() as f32 * 1e-6,
-                ))
-                .build(ui, &mut memory_length);
-            self.frame_view.lock().set_max_recent(memory_length as _);
-        }
+        max_memory_controls(ui, &frames, &self.frame_view);
 
         ui.columns(2, "columns", false);
         ui.set_column_width(0, 64.0);
@@ -701,7 +724,7 @@ impl ProfilerUi {
                     let duration = frame.duration_ns();
                     slowest_visible_frame = duration.max(slowest_visible_frame);
 
-                    let is_selected = Some(frame.frame_index) == selected_frame_index;
+                    let is_selected = Some(frame.frame_index()) == selected_frame_index;
 
                     let is_hovered = rect_min.x - 0.5 * frame_spacing <= mouse_pos.x
                         && mouse_pos.x < rect_max.x + 0.5 * frame_spacing
@@ -1204,3 +1227,39 @@ fn paint_thread_info(info: &Info<'_>, thread_info: &ThreadInfo, pos: [f32; 2]) {
 
     info.draw_list.add_text(pos, [0.9, 0.9, 0.9, 1.0], text);
 }
+
+fn max_memory_controls(ui: &Ui<'_>, frames: &Frames, frame_view: &GlobalFrameView) {
+    let uniq = frames.all_uniq();
+
+    let mut bytes = 0;
+    let mut unpacked = 0;
+    for frame in &uniq {
+        bytes += frame.bytes_of_ram_used();
+        unpacked += frame.has_unpacked() as usize;
+    }
+    ui.text(format!(
+        "{} frames ({} unpacked) using approximately {:.1} MB.",
+        uniq.len(),
+        unpacked,
+        bytes as f64 * 1e-6
+    ));
+
+    let frames_per_second = if let (Some(first), Some(last)) = (uniq.first(), uniq.last()) {
+        let nanos = last.range_ns().1 - first.range_ns().0;
+        let seconds = nanos as f64 * 1e-9;
+        let frames = last.frame_index() - first.frame_index() + 1;
+        frames as f64 / seconds
+    } else {
+        60.0
+    };
+
+    let mut memory_length = frame_view.lock().max_recent() as u64;
+    imgui::Slider::new("Max recent frames to store", 10, 100_000)
+        .display_format(format!(
+            "%d (~ {:.1} minutes, ~ {:.0} MB)",
+            memory_length as f64 / 60.0 / frames_per_second,
+            memory_length as f64 * bytes as f64 / uniq.len() as f64 * 1e-6,
+        ))
+        .build(ui, &mut memory_length);
+    frame_view.lock().set_max_recent(memory_length as _);
+}
